#ifndef REFLECTION_H
#define REFLECTION_H
#include "lrt.h"
#include "geometry.h"
class BSDF {
public:
	BSDF(const DifferentialGeometry &dg, BxDF *r, ...);
	~BSDF();
	int NumSpecular() const;
	Vector WorldToLocal(const Vector &v) const {
		return Vector(Dot(v, S), Dot(v, T), Dot(v, N)).Hat();
	}
	Vector LocalToWorld(const Vector &v) const {
		return Vector(S.x * v.x + T.x * v.y + N.x * v.z,
			      S.y * v.x + T.y * v.y + N.y * v.z,
			      S.z * v.x + T.z * v.y + N.z * v.z);
	}
	Spectrum f(const Vector &wiW, const Vector &woW) const;
	Spectrum f_delta(int component, const Vector &wi,
		Vector *wo) const;
	Spectrum sample_f(const Vector &wi, Vector *wo,	Float *weight) const {
		// XXX do a cosine-weighted hemisphere sample
		return 0.;
	}
	Float weight(const Vector &wi, const Vector &wo) const {
		// weight for above.
		return 0.;
	}
private:
	Vector N, S, T;
	vector<BxDF *> bxdfs;
	vector<Float> weights;
};
class BxDF {
public:
	virtual ~BxDF() { }
	virtual Spectrum f(const Vector &wi, const Vector &wo) const = 0;
	virtual bool IsSpecular() const { return false; }
	virtual Spectrum f_delta(const Vector &wi, Vector *wo) const {
		return 0.;
	}
	virtual Spectrum rho() const;
	virtual Spectrum rho(const Vector &w) const;
	static bool sameHemisphere(const Vector &wi, const Vector &wo) {
		return wi.z * wo.z > 0.;
	}
	virtual Spectrum sample_f(const Vector &wi, Vector *wo,
		Float *weight) const {
		// XXX do a cosine-weighted hemisphere sample
		return 0.;
	}
	virtual Float weight(const Vector &wi, const Vector &wo) const {
		// weight for above.
		return 0.;
	}
};
class LambertianReflection : public BxDF {
public:
	LambertianReflection(const Spectrum &reflectance) {
		R = reflectance;
	}
	Spectrum f(const Vector &wi, const Vector &wo) const;
	virtual Spectrum rho() const { return R; }
	virtual Spectrum rho(const Vector &w) const { return R; }
private:
	Spectrum R;
};
class LambertianTransmission : public BxDF {
public:
	LambertianTransmission(const Spectrum &tr) {
		T = tr;
	}
	Spectrum f(const Vector &wi, const Vector &wo) const;
	virtual Spectrum rho() const { return T; }
	virtual Spectrum rho(const Vector &w) const { return T; }
private:
	Spectrum T;
};
class MicrofacetDistribution {
public:
	virtual ~MicrofacetDistribution() { }
	virtual Float D(const Vector &wi, const Vector &wo) const = 0;
};

class Microfacet : public BxDF {
public:
	Microfacet(const Spectrum &reflectance, MicrofacetDistribution *d);
	~Microfacet();
	Spectrum f(const Vector &wi, const Vector &wo) const;
	Float G(const Vector &wi, const Vector &wo) const {
		return 1.f;
	}
	Float F(const Vector &wi, const Vector &wo) const {
		return 1.f;
	}
private:
	Spectrum R;
	MicrofacetDistribution *distribution;
};
class Blinn : public MicrofacetDistribution {
public:
	Blinn(Float e) { exponent = e; }
	Float D(const Vector &wi, const Vector &wo) const {
		Vector H = (wi + wo).Hat();
		Float costhetah = fabs(H.z);
		return pow(max(0.f, costhetah), exponent);
	}
private:
	Float exponent;
};
class SpecularReflection : public BxDF {
public:
	SpecularReflection(const Spectrum &r);
	Spectrum f(const Vector &, const Vector &) const {
		return Spectrum(0.);
	}
	bool IsSpecular() const { return true; }
	Spectrum f_delta(const Vector &wi, Vector *wo) const;
	Spectrum Sample(Float u[2], Vector *wi, Float *pdf) const
		{ *pdf = 0.; return Spectrum(0.); }
	Float Pdf(const Vector &wi) const { return 0.; }
private:
	Spectrum R;
};
class SpecularTransmission : public BxDF {
public:
	SpecularTransmission(const Spectrum &r, Float indexi, Float indext);
	Spectrum f(const Vector &, const Vector &) const {
		return Spectrum(0.);
	}
	bool IsSpecular() const { return true; }
	Spectrum f_delta(const Vector &wi, Vector *wo) const;
	Spectrum Sample(Float u[2], Vector *wi, Float *pdf) const
		{ *pdf = 0.; return Spectrum(0.); }
	Float Pdf(const Vector &wi) const { return 0.; }
private:
	Spectrum R;
	Float eta;
};
#endif // REFLECTION_H
